# -*- coding: utf-8 -*-
import io, ipaddress, sys, argparse, json, ast
from operator import itemgetter
from copy import deepcopy
from importlib.metadata import version
from datetime import datetime, timedelta
from a9x_webstatistics import __version__
from .gencockpitsubV0001 import *
    
def runGenCockpitV0001(infile, outfile, verbosity, domain, omit, efeature):

    owndomain = domain.replace('http://','')
    owndomain = owndomain.replace('https://','')
    owndomain = owndomain.replace('www.','')
    
    with open(infile) as json_file:
        d = json.load(json_file) 
        
        h = genHeader(owndomain)
        h += "\n" + '<div class="row pt-3"><div class="col-12">'
        h += '<h1>Webstatistics for ' + owndomain + '</h1>'
        h += '<a href="#hdDay" class="btn btn-primary me-3" role="button" data-bs-toggle="button">Daily Statistics</a>'
        h += '<a href="#hdLongTherm" class="btn btn-primary me-3" role="button" data-bs-toggle="button">Long Therm Statistics</a>'
        h += '</div></div>'
        h += "\n" + '<div class="row pt-4"><div class="col-12">'
        h += '<h2 id="hdDay">Daily statistics for ' + owndomain + '</h2>'
        h += "<p><small>Last record included in statistic: <script>document.write(DT2Locale('" + d['timelastrec'] + "'));</script></small></p>"
        h += '</div></div>'

        h += "\n" + '<div class="row pt-4">' + dailyHitsVisitsChart(d, owndomain, omit) + '</div>' + "\n"

        h += "\n" + '<div class="row pt-4">' + navChord(d, owndomain, omit) + '</div>' + "\n"

        # lastDate is the youngest date in the json
        # firstDate is the oldest date in the json

        lastDate = sorted(list(d["v0001"]["days"].keys() ))[-1]
        if verbosity == '99':
            print("gencockpit: lastDate (youngest) from json:" + str(lastDate) )
        if len(lastDate) >= 6:
            actYearMonth = lastDate[0:6]
        else:
            actYearMonth = lastDate[0:4] + '12'
            lastDate = lastDate[0:4] + '1231'

        firstDate = sorted(list(d["v0001"]["days"].keys() ))[0]
        if verbosity == '99':
            print("gencockpit: firstDate (oldest) from json:" + str(firstDate) )
        if len(firstDate) == 6:
            firstDate += '01'   # add day to YYYYMM
        if len(firstDate) == 4:
            firstDate += '0101'   # add MonthDay to YYYY
        d1 = datetime.strptime(firstDate, "%Y%m%d")
        
        if len(lastDate) == 6:
            if lastDate[4:6] == '02':
                lastDate += '28'   # add day to YYYYMM
            else:
                lastDate += '30'   # add day to YYYYMM
        if len(lastDate) == 4:
            lastDate += '1231'   # add MonthDay to YYYY
        d2 = datetime.strptime(lastDate, "%Y%m%d")

        # difference between dates in timedelta
        deltaInDays = d2 - d1
        if verbosity == '99':
            print("gencockpit: firstDate:" + str(firstDate) + " lastDate: " + str(lastDate) )
            print(f'Difference is {deltaInDays.days} days')

        # row for top domains and urls:
        h += "\n" + '<div class="row pt-4">'
        h += dailyTopDomains(d, owndomain, omit)
        h += dailyTopUrl(d, owndomain, omit)
        h += dailyTopCountries(d, owndomain, omit)
        h += '</div>'

        # accepted languages by browser:
        h += "\n" + '<div class="row pt-4">' + cockpitLanguages(d, owndomain) + '</div>' + "\n"

        # cache, response_time:
        h += "\n" + '<div class="row pt-4">' + performance(d) + '</div>'

        # quality:
        h += "\n" + '<div class="row pt-4">' + dailyQuality(d, owndomain, omit, verbosity) + '</div>'
        

        # show long term not on first runs:
        if deltaInDays.days > 32:

            h += "\n" + '<h2 id="hdLongTerm" class="pt-4">Long Term Statistics for ' + owndomain + '</h2>'
        
            h += "\n" + '<div class="row pt-4">' + monthlyHitsVisitsChart(d, owndomain, omit, verbosity) +  "\n"
            h +=  monthlyTopDomains(d, owndomain, omit) + '</div>' + "\n"
        
            h += "\n" + '<div class="row pt-4">' + monthlyTopUrl(d, owndomain, omit, verbosity) + "\n"
            h += monthlyTopCountries(d, owndomain, omit) + '</div>' + "\n"

            h += "\n" + '<div class="row pt-4">' + navChordLongterm(d, owndomain, omit) + '</div>' + "\n"
 
        h += "\n" + '<footer class="row bg-light py-4 mt-4">'
        h += '<div class="col"><h2>About</h2>'
        h += '<ul class="nav flex-column">'
        h += '<li class="nav-item mb-2"><a href="https://github.com/ava007/a9x-webstatistics" class="nav-link p-0 text-body-secondary">© 2024-2026 a9x-webstatistics</a></li>'
        h += '<li class="nav-item mb-2">' + __version__ + '</li>'
        h += '<li class="nav-item mb-2">salvo errore et omissione</li>'
        h += '</ul></div>' + "\n"

        h += '<div class="col"><h2>Sponsored by</h2>'
        h += '<ul class="nav flex-column">'
        h += '<li class="nav-item mb-2"><a href="https://www.logikfabrik.com" class="nav-link p-0 text-body-secondary">LogikFabrik.com - Knowledge Management</a></li>'
        h += '<li class="nav-item mb-2"><a href="https://www.logikfabrik.com/aboutwebstats.html" class="nav-link p-0 text-body-secondary">Live Examples</a></li>'
        h += '</ul></div>' + "\n"
          
        h += '<div class="col"><h2>Abbreviations</h2>'
        h += '<ul class="nav flex-column">'
        h += '<li class="nav-item mb-2"><a href="#" class="nav-link p-0 text-body-secondary">Hit: Download request of a html file</a></li>'
        h += '<li class="nav-item mb-2"><a href="#" class="nav-link p-0 text-body-secondary">URL: Uniform Resource Locator</a></li>'
        h += '</ul></div>' + "\n"

        h += '</footer>'

        h += '</div>'  # end of class "container"
        h += '</body></html>'

        # write html to file:
        outfile = open(outfile, "w")
        outfile.write(h)
        outfile.close()

        return 0
        
def is_valid_ip(address):
    try: 
        x = ipaddress.ip_address(address)
        return True
    except:
        return False
        
if __name__ == "__main__":
    parser = argparse.ArgumentParser(allow_abbrev=False,
        prog='a9x_webstatistics',
        epilog="Version: "+ version('a9x-webstatistics')
    )
    parser.add_argument("-i", "--infile", help="json file that contains calculated statistics", default="webstat.json")
    parser.add_argument("-o", "--outfile", help="html file that contains html cockpit", default="webstat.html")
    parser.add_argument("-d", "--domain", help="domain https://logikfabrik.com on which the access log file runs", default="https://logikfabrik.com")
    parser.add_argument("-t", "--omit", help="omits path ie. will not display these path in generated cockpit", action="append")
    parser.add_argument("-v", "--verbose", help="increase output verbosity, 0=none, 1=increased verbosity", default="0")
    parser.add_argument("-ef", "--efeature", help="use experimentalfeature number, 0=none, 1-99=feature", default="0")
    args, unknown = parser.parse_known_args()

    if args.omit is None:
        args.omit = []        

    runGenCockpitV0001(infile=args.infile, outfile=args.outfile, verbosity=args.verbose, domain=args.domain, omit=args.omit, efeature=args.efeature)



from .cockpitlanguages import cockpitLanguages
from .dailyHits import dailyHitsVisitsChart
from .dailytopdomains import dailyTopDomains
from .dailytopcountries import dailyTopCountries
from .dailytopurl import dailyTopUrl
from .dailyquality import dailyQuality
from .monthlyHits import monthlyHitsVisitsChart
from .externalfriends import externalFriends
from .genheader import genHeader
from .makenodeslinks import makeNodesLinks
from .makenodeslinks import makeNodesLinks2
from .monthlytopdomains import monthlyTopDomains
from .monthlytopurl import monthlyTopUrl
from .monthlytopcountries import monthlyTopCountries
from .navchart import navchart
from .navchartsankey import navchartsankey
from .navtangledtree import navchartTangledtree
from .navtangledtree2 import navchartTangledtree2
from .navChord import navChord
from .nav_chord_longterm import navChordLongterm
from .validip import is_valid_ip
from .performance import performance



from urllib.parse import urlparse
from operator import itemgetter
import ipaddress

def is_valid_ip(address):
    try: 
        x = ipaddress.ip_address(address)
        return True
    except:
        return False

def updV0001(
    d,
    i,
    visitIP,
    domain      # own domain
    ):

    # omit css and images
    if i['request'].endswith(('.css', '.ico', '.jpg', '.jpeg', '.png','.json','.txt')):
        return d, visitIP

    owndomain = domain.replace('http://','')
    owndomain = owndomain.replace('https://','')
    owndomain = owndomain.replace('www.','')
    owndomain = owndomain.lower()   # ensure that domain name is always in lowercase
        
    dt = i['ymd']
    req = urlparse(i['request'])

    # init a new day with minimal attributes:
    if 'v0001' not in d:
        d['v0001'] = {}
    if 'days' not in d['v0001']:
        d['v0001']['days'] = {}
    if dt not in d['v0001']['days']:
        d['v0001']['days'][dt] = {}
        d['v0001']['days'][dt]['user'] = {}
        d['v0001']['days'][dt]['user']['visits']  = 0
        d['v0001']['days'][dt]['user']['bytesSent'] = 0

        d['v0001']['days'][dt]['user']['deviceHits'] = {}
        d['v0001']['days'][dt]['user']['deviceHits']['mobile']  = 0 
        d['v0001']['days'][dt]['user']['deviceHits']['tablet']  = 0 
        d['v0001']['days'][dt]['user']['deviceHits']['desktop'] = 0 
        d['v0001']['days'][dt]['user']['serverResponseCode'] = {}

        d['v0001']['days'][dt]['robot'] = {}
        d['v0001']['days'][dt]['robot']['robotHits'] = 0
        d['v0001']['days'][dt]['robot']['bytesSent'] = 0
        d['v0001']['days'][dt]['robot']['serverResponseCode'] = {}
        d['v0001']['days'][dt]['quality'] = {}

    devCla = detectDeviceClass(i['user_agent']) 
    # update statistics for user:
    if devCla in ('desktop','mobile','tablet'):
        d['v0001']['days'][dt]['user']['bytesSent'] += int(i['bytes_sent'])

        if i['status'] in d['v0001']['days'][dt]['user']['serverResponseCode']:
            d['v0001']['days'][dt]['user']['serverResponseCode'][i['status']] += 1
        else:
            d['v0001']['days'][dt]['user']['serverResponseCode'][i['status']] = 1

        # process country if available:
        if 'country' in i:
            if 'countryHits' not in d['v0001']['days'][dt]['user']:
                d['v0001']['days'][dt]['user']['countryHits'] = {}
            if i['country'] not in d['v0001']['days'][dt]['user']['countryHits']:
                d['v0001']['days'][dt]['user']['countryHits'][i['country']] = 0
            d['v0001']['days'][dt]['user']['countryHits'][i['country']] += 1

        # update hits for device types:
        if devCla not in d['v0001']['days'][dt]['user']['deviceHits']:
            d['v0001']['days'][dt]['user']['deviceHits'][devCla] = 0 
        d['v0001']['days'][dt]['user']['deviceHits'][devCla] += 1 
                    
        d['timelastrec'] = i['timestamp']

        # update visits
        if i['ip'] not in visitIP :
            d['v0001']['days'][dt]['user']['visits'] = d['v0001']['days'][dt]['user']['visits'] + 1;
            visitIP[i['ip']] = 1

        # top urls: accumulate top urls - omitting query strings
        if i['status'] == '200':
            if 'topUrl' not in d['v0001']['days'][dt]['user']:
                d['v0001']['days'][dt]['user']['topUrl'] = {}
            if req.path not in d['v0001']['days'][dt]['user']['topUrl']:
                d['v0001']['days'][dt]['user']['topUrl'][req.path] = 0
            d['v0001']['days'][dt]['user']['topUrl'][req.path] += 1

        # update nav:
        if (i['status'] == '200' 
            and len(i['referer']) > 1 
            and devCla in ('desktop','mobile','tablet') ):
            refurl = urlparse(i['referer']).netloc
            if 'nav' not in d['v0001']['days'][dt]['user']:
                d['v0001']['days'][dt]['user']['nav'] = []
            tmpnav = {}
            tmpnav['s'] = None       # Source
            tmpnav['t'] = req.path  # Target
            tmpnav['c'] = 1         # count
            
            # external nav
            if (i['referer'][0:4] == 'http'):
                rdomain = refurl.removeprefix('www.')
                rdomain = rdomain.removesuffix(':80')    # to avoid duplicates: with or without ports
                # omit subdomains:
                if (rdomain.count('.') > 1 
                    and is_valid_ip(rdomain) == False):
                    rdomain_ar = rdomain.split('.')
                    rdomain = rdomain_ar[-2] + '.' + rdomain_ar[-1]
                tmpnav['s'] = rdomain.lower()   # source always in lowercase to avoid duplicates
                tmpnav['p'] = 'e'  # type: e=external source, i=internal source
            
            # internal nav
            else:
                #tmpnav['p'] = 'i'  # type: e=external source, i=internal source
                tmpnav['s'] = req.path   # Source

            if (tmpnav['s'] is not None 
                and tmpnav['t'] is not None):
                found = False        
                for e in d['v0001']['days'][dt]['user']['nav']:
                    if (e['s'] == tmpnav['s']
                        and e['t'] == tmpnav['t']):
                        e['c'] += tmpnav['c']
                        found = True
                        break
                if found == False:
                    d['v0001']['days'][dt]['user']['nav'].append(tmpnav)

    # update statistics for ROBOTS:
    if devCla not in ('desktop','mobile','tablet'):
        d['v0001']['days'][dt]['robot']['bytesSent'] += int(i['bytes_sent'])
        d['v0001']['days'][dt]['robot']['robotHits'] += 1
        
        if i['status'] in d['v0001']['days'][dt]['robot']['serverResponseCode']:
            d['v0001']['days'][dt]['robot']['serverResponseCode'][i['status']] += 1
        else:
            d['v0001']['days'][dt]['robot']['serverResponseCode'][i['status']] = 1
        
    # update quality: internal server error
    if i['status'] == '500':
        if 'quality' not in d['v0001']['days'][dt]:
            d['v0001']['days'][dt]['quality'] = {}
        if i['request'] not in d['v0001']['days'][dt]['quality']:
            d['v0001']['days'][dt]['quality'][i['request']] = {}
            d['v0001']['days'][dt]['quality'][i['request']]['from'] = i['referer']
            d['v0001']['days'][dt]['quality'][i['request']]['status'] = i['status']
            d['v0001']['days'][dt]['quality'][i['request']]['count'] = 1
            d['v0001']['days'][dt]['quality'][i['request']]['comment'] = 'internal server error'
        else:
            d['v0001']['days'][dt]['quality'][i['request']]['count'] += 1

    # update quality: internal permanently redirects: 301
    if (len(i['referer']) > 1
            and i['status'] == '301'
            and devCla in ('desktop','mobile','tablet')
            and owndomain in i['referer']):   # internal only
        if 'quality' not in d['v0001']['days'][dt]:
            d['v0001']['days'][dt]['quality'] = {}
        if i['request'] not in d['v0001']['days'][dt]['quality']:
            d['v0001']['days'][dt]['quality'][i['request']] = {}
            d['v0001']['days'][dt]['quality'][i['request']]['from'] = i['referer']
            d['v0001']['days'][dt]['quality'][i['request']]['status'] = i['status']
            d['v0001']['days'][dt]['quality'][i['request']]['count'] = 1
            d['v0001']['days'][dt]['quality'][i['request']]['comment'] = 'internal permanent redirect'
        else:
            d['v0001']['days'][dt]['quality'][i['request']]['count'] += 1

    # update quality: internal not found    
    if (len(i['referer']) > 1 
            and i['status'] == '404'
            and devCla in ('desktop','mobile','tablet')
            and owndomain in i['referer']):   # internal only
        if 'quality' not in d['v0001']['days'][dt]:
            d['v0001']['days'][dt]['quality'] = {}
        if i['request'] not in d['v0001']['days'][dt]['quality']:
            d['v0001']['days'][dt]['quality'][i['request']] = {}
            d['v0001']['days'][dt]['quality'][i['request']]['from'] = i['referer']
            d['v0001']['days'][dt]['quality'][i['request']]['status'] = i['status']
            d['v0001']['days'][dt]['quality'][i['request']]['count'] = 1
            d['v0001']['days'][dt]['quality'][i['request']]['comment'] = 'internal not found'
        else:
            d['v0001']['days'][dt]['quality'][i['request']]['count'] += 1
 

    # cache status for all successful requests:
    if ('cache_status' in i 
            and i['status'] == '200'):
        if len(i['cache_status']) == 0:  # avoid empty values
            i['cache_status'] = 'unknown'
        if 'performance' not in d['v0001']['days'][dt]:
           d['v0001']['days'][dt]['performance'] = {} 
        if 'cache' not in d['v0001']['days'][dt]['performance']:
           d['v0001']['days'][dt]['performance']['cache'] = {} 
        if i['cache_status'] not in d['v0001']['days'][dt]['performance']['cache']:
            d['v0001']['days'][dt]['performance']['cache'][i['cache_status']] = 1
        else:
            d['v0001']['days'][dt]['performance']['cache'][i['cache_status']] += 1

    # response time:
    if ('response_time' in i 
            and len(i['response_time']) > 0):
        tmp_rtime = float(i['response_time'])
        if tmp_rtime > 0:
            if 'performance' not in d['v0001']['days'][dt]:
               d['v0001']['days'][dt]['performance'] = {} 
            if 'response_time' not in d['v0001']['days'][dt]['performance']:
               d['v0001']['days'][dt]['performance']['response_time'] = {'time_count': 0, 'time_sum': 0} 
            d['v0001']['days'][dt]['performance']['response_time']['time_sum'] += tmp_rtime
            d['v0001']['days'][dt]['performance']['response_time']['time_count'] += 1

            # top x fastest:
            if 'topFast' not in d['v0001']['days'][dt]['performance']:
                d['v0001']['days'][dt]['performance']['topFast'] = {} 
                d['v0001']['days'][dt]['performance']['topFast'][i['request']] = tmp_rtime
            else:
                max_value = float(max(d['v0001']['days'][dt]['performance']['topFast'].values()))  # maximum value
                if tmp_rtime  < max_value:
                    d['v0001']['days'][dt]['performance']['topFast'][i['request']] = tmp_rtime

                if len(d['v0001']['days'][dt]['performance']['topFast']) > 5:
                    tmpFast = dict(sorted(d['v0001']['days'][dt]['performance']['topFast'].items(), key=lambda item: item[1], reverse=False))
                    z = 0
                    d['v0001']['days'][dt]['performance']['topFast'] = {}
                    for k,v in tmpFast.items():
                        z += 1
                        if z >= 5:
                            break
                        d['v0001']['days'][dt]['performance']['topFast'][k] = v
                    del tmpFast

            # top Slowest:
            if 'topSlow' not in d['v0001']['days'][dt]['performance']:
                d['v0001']['days'][dt]['performance']['topSlow'] = {} 
                d['v0001']['days'][dt]['performance']['topSlow'][i['request']] = tmp_rtime
            else:
                min_value = float(min(d['v0001']['days'][dt]['performance']['topSlow'].values()))  # minimum value
                if tmp_rtime > min_value:
                    d['v0001']['days'][dt]['performance']['topSlow'][i['request']] = tmp_rtime

            if len(d['v0001']['days'][dt]['performance']['topSlow']) > 5:
                tmpSlow = dict(sorted(d['v0001']['days'][dt]['performance']['topSlow'].items(), key=lambda item: item[1], reverse=True))
                z = 0
                d['v0001']['days'][dt]['performance']['topSlow'] = {}
                for k,v in tmpSlow.items():
                    z += 1
                    if z >= 5:
                        break
                    d['v0001']['days'][dt]['performance']['topSlow'][k] = v
                del tmpSlow
    
    # accept language for user hits only:
    # ru-RU,ru;q=0.9,en-US;q=0.8,en;q=0.7
    # en
    # en-US,en;q=0.5
    # de,en-US;q=0.7,en;q=0.3
    # zh-CN,zh;q=0.9,en-US;q=0.8,en;q=0.7
    tmp_lang = []
    if i['status'] == '200' and 'accept_language' in i and len(i['accept_language']) > 0 and devCla in ('desktop','mobile','tablet'):
        if 'language' not in d['v0001']['days'][dt]['user']:
            d['v0001']['days'][dt]['user']['language'] = {}
        for k,v in i['accept_language'].items():
            lng = k[0:2]  # take language only, ignore country
            # count each language only once per hit:
            if lng not in tmp_lang:
                tmp_lang.append(lng)
                if lng not in d['v0001']['days'][dt]['user']['language']:
                    d['v0001']['days'][dt]['user']['language'][lng] = 0
                d['v0001']['days'][dt]['user']['language'][lng] += 1
    #print("\nd: " + str(dt) + " "  + str(d['v0001']['days'][dt]) ) 
                
    return d, visitIP

def detectDeviceClass(ua):
  if ua is None:
     return 'others'
  if (ua.lower().find('bot') > 0):
     return 'bots'
  if (ua.lower().find('externalhit') > 0):    # facebook
     return 'bots'
  if (ua.lower().find('spider') > 0):        # bytedance
     return 'bots'
  if (ua.lower().find('crawler') > 0):
     return 'others'
  if (ua.lower().find('python') > 0):        # python-requests
     return 'bots'
  if (ua.lower().find('mediapartner') > 0):
     return 'others'
  if (ua.lower().find('feedly') > 0):        # Feedly
     return 'bots'
  if (ua.lower().find('validator') > 0):     # validator
     return 'bots'
  if (ua.lower().find('curl') > 0):
     return 'others'
  if (ua.lower().find('ipad') > 0):
     return 'tablet'
  if (ua.lower().find('mobile') > 0):
     return 'mobile'
  if (ua.lower().find('android') > 0):
     return 'tablet'
  return 'desktop'



# -*- coding: utf-8 -*-
from operator import itemgetter

def dailyQuality(d, owndomain, omit, verbosity):
    tquality = {}   # nested dictionary!
    daycount = 0
    h = ''
    for k, v in sorted(d['v0001']['days'].items(), key=itemgetter(0), reverse=True):
        if daycount > 10:
            break
        daycount += 1
        if 'quality' in d['v0001']['days'][k]:
            for sk,sv in d['v0001']['days'][k]['quality'].items():
                #print('sk: ' + str(sk) + ' sv: ' + str(sv))
                if sk not in tquality:
                    tquality[sk] = {}
                    tquality[sk]['count'] = 1
                    tquality[sk]['status'] = sv['status']
                    if sv['status'][0:1] == '5':  
                        tquality[sk]['prio'] = 'A'
                    elif sv['status'][0:1] == '4':
                        tquality[sk]['prio'] = 'B'
                    else:
                        tquality[sk]['prio'] = 'Z'
                    tquality[sk]['from'] = sv['from']
                    tquality[sk]['comment'] = sv['comment']
                    tquality[sk]['lastoccured'] = k
                else:
                    tquality[sk]['count']  += 1

    if len(tquality) > 0:
        h += '<div class="col-md-12 col-lg-12 col-xxl-12">'
        h += '<h3>Quality Improvements</h3>'
        h += '<p class="card-text">Quality improvements that should be implemented:</p>'
        h += '<table class="table-responsive text-break"><thead><tr><th scope="col">Status</th><th scope="col">affected URL</th><th scope="col">Count</th><th scope="col">Remark</th><th scope="col">Date last occured</th></tr></thead>'
        i = 0
        for k, v in sorted(tquality.items(), key=lambda x: (x[1]['prio'])):
            h += '<tr><td>' + str(v['status']) + '</td><td>' + str(k) + "</td>" 
            h += '<td style="text-align: right">' + str(format(v['count'],',')) + '</td>'
            h += '<td>' + v['comment'] + '</td>'
            h += "<td><script>document.write(DT2Locale('" + v['lastoccured'] + "'));</script></td>"
            h += '</tr>'
            if len(v['from']) > 1:
                h += '<tr><td></td><td colspan="4">called by: ' + v['from'] + '</td></tr>'
            i += 1
            if i == 10:
                break
        h += '</table>'
        h += '</div>'   # end of card and col
 
    return h



import argparse
import io
import json
import re
import sys
from datetime import datetime
from .updatestatistics import upd
from .updateStatisticsV0001 import updV0001
#from .summarizemonth import summonth
#from .summarizemonthV0001 import sumMonthV0001
from .migV0001 import migv0001sub0001
from .cleanupV0001 import cleanupQualityV0001
from .parseRecJsonV0001 import parseRecJsonV0001
from .sumDayV0001 import sumDayV0001
from importlib.metadata import version
import geoip2.database
from .sumV0001 import *

def parseRec(rec, log_pattern, j, georeader):
    r = {}
    # skip comments recs
    if rec[0:1] == '#':
        j['records_skipped_comment'] += 1
        return r,j

    if rec[0:2] == '{"':
        r = parseRecJsonV0001(rec, georeader)
    else:
        data = re.search(log_pattern, rec)

        if data:
            datadict = data.groupdict()
            ip_address = datadict["ipaddress"]
            timestamp = datadict["dateandtime"]
            request = datadict["url"]
            bytes_sent = datadict["bytessent"]
            referer = datadict["referer"]
            user_agent = datadict["useragent"]
            status = datadict["statuscode"]
            method = datadict["method"]
            if georeader:
                try:
                    grrsp = georeader.country(ip_address)
                    country = grrsp.country.name
                except geoip2.errors.AddressNotFoundError:
                    country = None
            else:
                country = None
        
            dto = datetime.strptime(timestamp,'%d/%b/%Y:%H:%M:%S %z')  # 07/Jan/2024:14:06:24 +0000
                
            r = {
                'ip': ip_address,
                'ymd': dto.strftime("%Y%m%d"),
                'timestamp': dto.strftime("%Y%m%d%H%M%S") ,
                'request': request,
                'status': status,
                'bytes_sent': 0,
                'referer': referer,
                'user_agent': user_agent
            }
            if country:
                r['country'] = country
            if bytes_sent.isdigit():                 # bytes_sent can also hold "-"
                r['bytes_sent'] = int(bytes_sent)
        else:
            print("Log Rec parsing failed for: " + rec[0:79])
            j['records_parsing_failed'] += 1
        
    return r,j


def runws(statfile, infile, geoip, verbosity, domain, efeature = 0):

    try: 
        georeader = geoip2.database.Reader(geoip)
    except (FileNotFoundError, TypeError) as gerr:
        georeader = None
        print("geoip2 file not found, continue processing")

    # init statistic file if it does not exist:
    d = {}
    d['timelastrec'] = '19991231235959'
    d['v0001'] = {}
    d['v0001']['days'] = {}

    # init job results
    j = {
        'records_read': 0,
        'records_skipped_comment': 0,
        'records_parsing_failed': 0,
        'records_already_processed': 0,
        'records_processed_for_statistic': 0,
    }
    
    # load statistic file if it exists
    try:
        f = open(statfile) 
        d = json.load(f) 
    except FileNotFoundError:   # first call: file does not exists
        print("-s statistic file not found, it will be automatically created")
    except json.JSONDecodeError:
        print("-s json file is not valid")

    visitIP = {}
    lasttimerecobj = datetime.strptime(d['timelastrec'],"%Y%m%d%H%M%S")
    print("lasttimerecobj: " + str(lasttimerecobj))

    log_pattern = re.compile(
        r"""
        (?P<ipaddress>
        \d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|               # ipv4
        (                                                  
        ([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|          # ipv6 1:2:3:4:5:6:7:8
        ([0-9a-fA-F]{1,4}:){1,7}:|                         # ipv6 1::              1:2:3:4:5:6:7::
        ([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|         # ipv6 1::8             1:2:3:4:5:6::8  1:2:3:4:5:6::8
        ([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|  # ipv6 1::7:8           1:2:3:4:5::7:8  1:2:3:4:5::8
        ([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|  # ipv6 1::6:7:8         1:2:3:4::6:7:8  1:2:3:4::8
        ([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|  # ipv6 1::5:6:7:8       1:2:3::5:6:7:8  1:2:3::8
        ([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|  # ipv6 1::4:5:6:7:8     1:2::4:5:6:7:8  1:2::8
        [0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|       # ipv6 1::3:4:5:6:7:8   1::3:4:5:6:7:8  1::8  
        :((:[0-9a-fA-F]{1,4}){1,7}|:)|                     # ipv6 ::2:3:4:5:6:7:8  ::2:3:4:5:6:7:8 ::8       ::     
        fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|     # fe80::7:8%eth0   fe80::7:8%1  (link-local IPv6 addresses with zone index)
        ::(ffff(:0{1,4}){0,1}:){0,1}
        ((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}
        (25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|          # ::255.255.255.255   ::ffff:255.255.255.255  ::ffff:0:255.255.255.255  (IPv4-mapped IPv6 addresses and IPv4-translated addresses)
        ([0-9a-fA-F]{1,4}:){1,4}:
        ((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}
        (25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])           # 2001:db8:3:4::192.0.2.33  64:ff9b::192.0.2.33 (IPv4-Embedded IPv6 Address)
        )
        )[ ]-[ ](?P<uid>.+)[ ]\[(?P<dateandtime>\d{2}\/[a-zA-Z]{3}\/\d{4}:\d{2}:\d{2}:\d{2}[ ](\+|\-)\d{4})\][ ]
        ((?P<method>)(\"(GET|POST|HEAD|PUT|DELETE|OPTIONS|PROPFIND))[ ]
        (?P<url>.+)[ ](HTTP\/(1\.0|1\.1|2\.0)"))[ ]
        (?P<statuscode>\d{3})[ ]
        (?P<bytessent>\S+)[ ]                              # number, can also be "-"
        (["](?P<referer>[^"]+)["])[ ]
        (["](?P<useragent>[^"]+)["])
        """,
        flags=re.VERBOSE
    )
    
    # process infile:
    with open(infile,'r') as infile:
        for rec in infile:
            j['records_read'] += 1
            recparsed, j = parseRec(rec, log_pattern, j, georeader)
            # skip unrecognized records:
            if not recparsed or recparsed['timestamp'] is None or recparsed['ip'] is None:
                continue
            # skip already processed data:
            if recparsed['timestamp']  <=  d['timelastrec']:
                j['records_already_processed'] += 1
                continue
            d, visitIP = updV0001(d, recparsed, visitIP, domain)
            j['records_processed_for_statistic'] += 1
        if verbosity == "99":
            print("main: last recparsed:" + str(recparsed) )
 
    # sum navigation
    d = sumDayV0001(d, verbosity)
    
    # summarize previous months
    d = sumDay2MonthV0001(d, statfile, verbosity)

    # sum Months to Years:
    d = sumMonth2YearV0001(d, statfile, verbosity)

    d = cleanupQualityV0001(d)
     
    # write updated statistic file:
    with open(statfile, "w") as sf:
       json.dump(d,sf)  

    print("Rec read:           "      + str(j['records_read']))
    print("Rec processed:      " + str(j['records_processed_for_statistic']))
    print("Rec parsing failed: " + str(j['records_parsing_failed']))
    print("Rec skipped (already processed): "   + str(j['records_already_processed']))
    return 0

if __name__ == "__main__":
    parser = argparse.ArgumentParser(allow_abbrev=False,
        prog='a9x_webstatistics',
        epilog="Version: "+ version('a9x-webstatistics')
    )
    parser.add_argument("-s", "--statfile", help="json file that contains calculated statistics", default="webstat.json")
    parser.add_argument("-i", "--infile", help="filename including path to web server access log that contains input data", default="nginx_access.log")
    parser.add_argument("-g", "--geoip", help="path/file to GeoIP2-Country.mmdb file")
    parser.add_argument("-v", "--verbose", help="increase output verbosity, 0=none, 1=increased verbosity", default="0")
    parser.add_argument("-d", "--domain", help="domain https://logikfabrik.com on which the access log file runs", default="https://logikfabrik.com")
    parser.add_argument("-ef", "--efeature", help="use experimentalfeature number, 0=none, 1-99=feature", default="0")
    args, unknown = parser.parse_known_args()
    
    #migv0001sub0001(args.statfile)
    runws(statfile=args.statfile, infile=args.infile, geoip=args.geoip, verbosity=args.verbose, domain=args.domain, efeature=args.efeature)
